@@ -27,6 +27,7 @@
 #     those of the authors and should not be interpreted as representing
 #     official policies, either expressed or implied, of the FreeBSD Project.
 
+from __future__ import print_function
 
 import operator
 from functools import wraps
@@ -51,8 +52,33 @@ def Config(filename, dictionary=None, overwrite=True):
     return dictionary
 
 
+def _RangeCaster(value):
+    def raise_error():
+        raise FormatError(('%s is not range format: N:N+i, N-N+i, or N'
+                + ' N+i') % value)
+
+    splitter = None
+    for char in (' :-'):
+        if char in value:
+            splitter = char
+            break
+
+    if splitter:
+        toks = value.split(splitter)
+    else:
+        toks = [value]
+
+    if not (1 <= len(toks) <= 3):
+        raise_error()
+    try:
+        return xrange(*[int(y) for y in toks])
+    except ValueError:
+        raise_error()
+
+
 # ---------- decorators ---------- #
 
+
 def _names_to_options(f):
     @wraps(f)
     def inner(*args, **kwargs):
@@ -183,7 +209,8 @@ class ManyAllowedNoneSpecifiedArgumentError(ArgumentError):
 class UnspecifiedArgumentError(ArgumentError):
     ''' User supplies argument that isn't specified. '''
 
-    pass
+    def __init__(self, arg):
+        super(UnspecifiedArgumentError, self).__init__('illegal option %s' % arg)
 
 
 class MultipleSpecifiedArgumentError(ArgumentError):
@@ -195,7 +222,8 @@ class DependencyError(ArgumentError):
     ''' User specified an argument that requires another, unspecified argument.
     '''
 
-    pass
+    def __init__(self, arg1, arg2):
+        super(DependencyError, self).__init__('%s requires %s' % (arg1, arg2))
 
 
 class ConflictError(ArgumentError):
@@ -333,6 +361,28 @@ class Option(Condition):
         [self._parser._set_conflicts(self.argname, x) for x in others]
         return self
 
+    def _cast(self):
+        return self._parser._casts.get(self.argname)
+
+    def _isrequired(self):
+        return self in self._parser._required
+
+    def _getconflicts(self):
+        return self._parser._conflicts.get(self, [])
+
+    def _getreqs(self):
+        return self._parser._requires.get(self, [])
+
+    def _alias(self):
+        return self._parser._source_to_alias.get(self.argname)
+
+    def __str__(self):
+        v = '%s%s' % (self._parser._double_prefix, self.argname)
+        alias = self._alias()
+        if alias:
+            v += '/%s%s' % (self._parser._single_prefix, alias)
+        return v
+
     def shorthand(self, alias):
         ''' Set shorthand for this argument. Shorthand arguments are 1
         character in length and are prefixed by a single '-'. For example:
@@ -387,7 +437,8 @@ class Option(Condition):
         return self
 
     def _inner_satisfied(self, parsed):
-        return self.argname in parsed
+        v = parsed.get(self.argname)
+        return v is not None and v != False
 
     def _make_condition(self, func, other):
         return CallableCondition(func, self, other)
@@ -508,6 +559,9 @@ class Group(Option):
                 return True
         return False
 
+    def __str__(self):
+        return ', '.join([str(name) for name in self._names])
+
 
 class Parser(object):
     ''' Command line parser. '''
@@ -532,7 +586,9 @@ class Parser(object):
 
         # dict of A -> args that A conflicts with
         self._conflicts = {}
+
         self._alias = {}
+        self._source_to_alias = {}
 
         # convert arg labels to understore in value dict
         self._to_underscore = False
@@ -546,6 +602,8 @@ class Parser(object):
         # help message
         self._help_prefix = None
 
+        self._out = sys.stdout
+
         # set by user
         self._init_user_set(store)
 
@@ -553,7 +611,6 @@ class Parser(object):
             self.flag('help').shorthand('h')
 
     def _init_user_set(self, store=None):
-        self._preparsed = {}
         if store is None:
             store = {}
         self._store = store
@@ -579,7 +636,7 @@ class Parser(object):
         (e.g., -a). '''
 
         if self._double_prefix in flag:
-            raise ValueError('single_flag cannot be superset of double_flag')
+            raise ValueError('single_prefix cannot be superset of double_prefix')
         self._single_prefix = flag
         return self
 
@@ -604,14 +661,6 @@ class Parser(object):
         vals = inspect.currentframe().f_back.f_locals
         return Parser(vals).underscore()
 
-    def _to_flag(self, name):
-        name = self._unlocalize(name)
-
-        if len(name) == 1:
-            return self._single_prefix + name
-        else:
-            return self._double_prefix + name
-
 # --- types --- #
 
     def config(self, name):
@@ -657,30 +706,8 @@ class Parser(object):
               python test.py --values 0 10 3  # -> xrange(0, 10, 3)
         '''
 
-        def caster(x):
-            def raise_error():
-                raise FormatError(('%s is not range format: N:N+i, N-N+i, or N'
-                        + ' N+i') % x)
 
-            splitter = None
-            for char in (' :-'):
-                if char in x:
-                    splitter = char
-                    break
-
-            if splitter:
-                toks = x.split(splitter)
-            else:
-                toks = [x]
-
-            if not (1 <= len(toks) <= 3):
-                raise_error()
-            try:
-                return xrange(*[int(y) for y in toks])
-            except ValueError:
-                raise_error()
-
-        result = self._add_option(name).cast(caster)
+        result = self._add_option(name).cast(_RangeCaster)
         self._set_reader(name, _MultiWordArgumentReader)
         return result
 
@@ -787,6 +814,7 @@ class Parser(object):
                 self._alias[alias]))
 
         self._alias[alias] = source
+        self._source_to_alias[source] = alias
 
     def _add_option(self, name, argument_label=None):
         name = self._localize(name)
@@ -828,7 +856,7 @@ class Parser(object):
             return key
         return v
 
-    def _parse(self, args):
+    def _tokenize(self, args):
         new_args = []
         for arg in args:
             if '=' in arg:
@@ -842,10 +870,11 @@ class Parser(object):
         return (arg.startswith(self._single_prefix) or
                 arg.startswith(self._double_prefix))
 
-    def _read(self, args):
+    def _parse(self, tokenized):
         argument_value = None
+        parsed = {}
 
-        for arg in args:
+        for arg in tokenized:
             if argument_value is not None:
                 if argument_value.consume_or_skip(arg):
                     continue
@@ -854,7 +883,9 @@ class Parser(object):
             argument_name = None
 
             if self._is_argument_label(arg):
-                if arg.startswith(self._double_prefix):
+                is_full = arg.startswith(self._double_prefix)
+
+                if is_full:
                     prefix = self._double_prefix
                 else:
                     prefix = self._single_prefix
@@ -863,10 +894,18 @@ class Parser(object):
 
                 argument_name = self._localize(arg)
 
-                argument_name, argument_value = self._getoption(argument_name)
+                if is_full:
+                    argument_name = self._options.get(argument_name).argname
+                    argument_value = self._readers.get(argument_name)
+                else:
+                    argument_name = self._options.get(self._alias.get(argument_name)).argname
+                    argument_value = self._readers.get(argument_name)
+
                 if argument_value is None:
                     raise UnspecifiedArgumentError(argument_name)
+
                 argument_value = argument_value(self)
+
             elif self._unspecified_default is not None:
                 argument_name = self._unspecified_default
 
@@ -875,19 +914,21 @@ class Parser(object):
                 argument_value.consume_or_skip(arg)
 
             if argument_name:
-                self._preparsed.setdefault(argument_name,
+                parsed.setdefault(argument_name,
                         []).append(argument_value)
             else:
                 self._extras.append(arg)
 
-    def _help_if_necessary(self):
-        if 'help' in self._preparsed:
+        return parsed
+
+    def _help_if_necessary(self, processed):
+        if 'help' in processed:
             self.print_help()
             sys.exit(0)
 
-    def _assign(self):
-        parsed = {}
-        for key, values in self._preparsed.iteritems():
+    def _assign(self, parsed):
+        assigned = {}
+        for key, values in parsed.iteritems():
             try:
                 if key not in self._multiple:
                     value = values[0].get()
@@ -897,40 +938,59 @@ class Parser(object):
                 cast = self._casts.get(key)
                 if cast:
                     try:
-                        value = cast(value)
+                        if isinstance(value, list):
+                            newv = []
+                            values = value
+                            for value in values:
+                                newv.append(cast(value))
+                            value = newv
+                        else:
+                            value = cast(value)
                     except FormatError:
                         raise
-                    except ValueError:
-                        raise FormatError('Cannot cast %s to %s', value, cast)
+                    except ValueError as e:
+                        raise FormatError('Cannot cast \'%s\' to %s' % (value, cast))
 
-                parsed[key] = value
+                assigned[key] = value
             except MissingValueError:
                 raise MissingValueError('%s specified but missing given value'
                         % key)
 
-        # check multiple
+        # assign defaults
+        for key, value in self._readers.iteritems():
+            if key in parsed:
+                continue
 
-        for key, values in self._preparsed.iteritems():
+            if key in self._defaults:
+                assigned[key] = self._defaults[key]
+            else:
+                assigned[key] = value.default()
+
+        return assigned
+
+    def _check_multiple(self, assigned):
+        for key, values in assigned.iteritems():
             if len(values) > 1 and key not in self._multiple:
-                raise MultipleSpecifiedArgumentError(('%s specified multiple'
-                        + ' times') % self._to_flag(key))
+                raise MultipleSpecifiedArgumentError(('%s specified multiple' +
+                    ' times') % self._options[key])
 
-        # check conditions
-        for arg in parsed.iterkeys():
+    def _check_conditions(self, assigned):
+        for arg in assigned.iterkeys():
             for cond in self._options[arg]._conditions:
-                if not cond(parsed):
+                if not cond(assigned):
                     raise FailedConditionError()
 
+    def _check_required(self, assigned):
         for arg, replacements in self._required.iteritems():
             missing = []
-            if not arg._is_satisfied(parsed):
+            if not arg._is_satisfied(assigned):
                 for v in replacements:
                     if isinstance(v, Group):
                         missing += v._names
                     elif not isinstance(v, Condition):
                         missing.append(v)
 
-                    if v._is_satisfied(parsed):
+                    if v._is_satisfied(assigned):
                         break
                 else:
                     if missing:
@@ -939,28 +999,31 @@ class Parser(object):
                     else:
                         raise MissingRequiredArgumentError(arg)
 
+    def _check_dependencies(self, assigned):
         for arg, deps in self._requires.iteritems():
-            if arg._is_satisfied(parsed):
+            if arg._is_satisfied(assigned):
                 for v in deps:
-                    if not v._is_satisfied(parsed):
+                    if not v._is_satisfied(assigned):
                         if isinstance(v, CallableCondition):
                             raise ConditionError(arg.argname, v)
-                        raise DependencyError(v)
+                        raise DependencyError(arg, v)
 
+    def _check_conflicts(self, assigned):
         for arg, conflicts in self._conflicts.iteritems():
-            if arg._is_satisfied(parsed):
+            if arg._is_satisfied(assigned):
                 for conflict in conflicts:
-                    if conflict._is_satisfied(parsed):
+                    if conflict._is_satisfied(assigned):
                         raise ConflictError(arg.argname, conflict.argname)
 
-        for key, value in parsed.iteritems():
-            self._set_final(key, value)
+    def _verify(self, assigned):
+        self._check_conditions(assigned)
+        self._check_required(assigned)
+        self._check_dependencies(assigned)
+        self._check_conflicts(assigned)
 
-        for key, value in self._readers.iteritems():
-            if key in self._defaults:
-                self._set_final(key, self._defaults[key])
-            else:
-                self._set_final(key, value.default())
+    def _assign_to_store(self, assigned):
+        for key, value in assigned.iteritems():
+            self._store[key] = value
 
     @_options_to_names
     def _set_one_required(self, *names):
@@ -979,18 +1042,22 @@ class Parser(object):
     def _process_command_line(self, args=None):
         try:
             args = self._get_args(args)
-            args = self._parse(args)
-            self._read(args)
-            self._help_if_necessary()
-
-            self._assign()
-            store = self._store
+            tokenized = self._tokenize(args)
+            parsed = self._parse(tokenized)
+            self._help_if_necessary(parsed)
+            self._check_multiple(parsed)
+            assigned = self._assign(parsed)
+            self._verify(assigned)
+            self._assign_to_store(assigned)
         except ArgumentError as e:
             raise e
-        finally:
-            self._init_user_set()  # reset
 
-        return store
+#        self._init_user_set()  # reset
+
+        return self._store
+
+    def emit(self, *args):
+        print(*args, file=self._out)
 
     def process_command_line(self, args=None):
         try:
@@ -1000,19 +1067,12 @@ class Parser(object):
 
     def bail(self, e):
         msg = []
-        if isinstance(e, UnspecifiedArgumentError):
-            msg.append('illegal option --%s' % e)
-        else:
-            msg.append(str(e))
+        msg.append('Error: ' + str(e))
         msg.append('usage: %s' % sys.argv[0])
-        msg.append(' '.join('[%s]' % self._label(key) for key in self._readers))
-        print('\n'.join(msg))
+        msg.append(' '.join('[%s]' % self._label(value) for value in self._options.values()))
+        self.emit('\n'.join(msg))
         sys.exit(1)
 
-    def _set_final(self, key, value):
-        if key not in self._store:
-            self._store[key] = value
-
     @localize
     @_options_to_names
     def _set_multiple(self, name):
@@ -1044,37 +1104,66 @@ class Parser(object):
     def __enter__(self):
         return self
 
-    def __exit__(self, a, b, c):
-        self.process_command_line()
+    def __exit__(self, exc_type, exc_value, traceback):
+        if exc_type is None or isinstance(exc_type, ArgumentError):
+            self.process_command_line()
+            return True
+        return False
 
-    def _option_label(self, name):
-        return self._option_labels.get(name, 'option')
+    def _option_label(self, name, opt):
+        label = self._option_labels.get(name)
+        if label is not None:
+            return label
 
-    def _label(self, name):
-        pkey = self._to_flag(name)
-        for k, v in self._alias.iteritems():
-            if name in v:
-                pkey += ',' + self._to_flag(k)
+        if opt._cast() is int:
+            return 'int'
 
-        value = self._readers[name]
-        if value != _FlagArgumentReader:
-            pkey = '%s <%s>' % (pkey, self._option_label(name))
+        if opt._cast() is float:
+            return 'float'
+
+        if opt._cast() is float:
+            return 'float'
+
+        if opt._cast() is _RangeCaster:
+            return 'range'
+
+        return 'option'
+
+    def _label(self, opt):
+        pkey = str(opt)
+
+        reader = self._readers[opt.argname]
+        if reader != _FlagArgumentReader:
+            pkey = '%s <%s>' % (pkey, self._option_label(opt.argname, opt))
 
         return pkey
 
     def print_help(self):
         if self._help_prefix:
-            print(self._help_prefix)
+            self.emit(self._help_prefix)
 
-        print('Arguments:')
+        self.emit('Arguments: (! denotes required argument)')
         column_width = -1
-        for key in self._readers:
-            column_width = max(column_width, len(self._label(key)))
+        for key, value in self._options.iteritems():
+            column_width = max(column_width, len(self._label(value)))
 
         fmt = '   %-' + str(column_width) + 's'
 
-        for key in self._readers:
-            print(fmt % self._label(key))
+        for key, value in self._options.iteritems():
+            msg = fmt % self._label(value)
+
+            if value._isrequired():
+                msg = '!' + msg[1:]
+
+            conflicts = value._getconflicts()
+            if conflicts:
+                msg += ' (Conflicts with %s)' % ', '.join(str(item) for item in conflicts)
+
+            reqs = value._getreqs()
+            if reqs:
+                msg += ' (Requires %s)' % ', '.join(str(item) for item in reqs)
+
+            self.emit(msg)
 
 
 class IOParser(Parser):
@@ -1125,4 +1214,5 @@ class IOParser(Parser):
         return self._add_option(name).cast(opener)
 
 
-__version__ = '0.2.12a'
+__all__ = ['Parser']
+__version__ = '0.2.14a'
